#include "data.lua"

--General functions
local function toEuler(quat)
    local x, y, z, w = quat[1], quat[2], quat[3], quat[4]
    -- Credit to https://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm

    local bank, heading, attitude

    local s = 2 * x * y + 2 * z * w
    if s >= 1 then
        heading = 2 * math.atan2(x, w)
        bank = 0
        attitude = math.pi / 2
    elseif s <= -1 then
        heading = -2 * math.atan2(x, w)
        bank = 0
        attitude = math.pi / -2
    else
        bank = math.atan2(2 * x * w - 2 * y * z, 1 - 2 * x ^ 2 - 2 * z ^ 2)
        heading = math.atan2(2 * y * w - 2 * x * z,
                             1 - 2 * y ^ 2 - 2 * z ^ 2)
        attitude = math.asin(s)
    end

    return math.deg(bank), math.deg(heading), math.deg(attitude)
end

local function VecAngle(v1, v2)
    local numerator = VecDot(v1, v2)
    local denominator = math.sqrt((v1[1] * v1[1]) + (v1[2] * v1[2]) + (v1[3] * v1[3])) * math.sqrt((v2[1] * v2[1]) + (v2[2] * v2[2]) + (v2[3] * v2[3]))
    local rad = math.acos(numerator/denominator)

    return math.deg(rad)
end

local function getPlayerDirection()
    local direction = {}

    local ct = GetPlayerCameraTransform()
    local pVec = TransformToParentVec(ct, Vec(0, 0, -1))

    local xzVec = Vec(pVec[1], 0, pVec[3])
    local xBaseVec = Vec(1, 0, 0)

    direction.yaw = VecAngle(xBaseVec, xzVec)
    direction.pitch = VecAngle(xzVec, pVec)

    if(pVec[1] < 0 and pVec[3] < 0 or pVec[1] > 0 and pVec[3] < 0) then
        direction.yaw = (180 - direction.yaw) + 180
    end 

    if(pVec[2] < 0) then
        direction.pitch = -direction.pitch
    end 

    return direction
end

local function getCameraFront()
    local direction = Vec(0, 0, 0)

    yaw = getPlayerDirection().yaw
    pitch = getPlayerDirection().pitch

    direction[1] = math.cos(math.rad(yaw)) * math.cos(math.rad(pitch))
    direction[2] = math.sin(math.rad(pitch))
    direction[3] = math.sin(math.rad(yaw)) * math.cos(math.rad(pitch))

    return VecNormalize(direction)
end

local function has_value(tab, val)
    for index, value in ipairs(tab) do
        if value == val then
            return true
        end
    end
    return false
end

local function progressBar(w, h, t) --common.lua line 44 (edited)
	UiPush()
		UiAlign("left top")
		UiColor(0, 0, 0, 0.5)
		UiImageBox("ui/common/box-solid-10.png", w, h, 6, 6)
		if t > 0 then
			UiTranslate(2, 2)
			w = (w-4)*t
			if w < 12 then w = 12 end
			h = h-4
            UiColor(1,(1-t),(1-(t*2)),1)
			UiImageBox("ui/common/box-solid-6.png", w, h, 6, 6)
		end
	UiPop()
end

local function round(x, n)
    n = math.pow(10, n or 0)
    x = x * n
    if x >= 0 then x = math.floor(x + 0.5) else x = math.ceil(x - 0.5) end
    return x / n
end

local function getBodySpeed(body)
    local trans = GetBodyTransform(body)
    local vel = TransformToLocalVec(trans, GetBodyVelocity(body))
    local speed = math.sqrt(vel[1]*vel[1]+vel[2]*vel[2]+vel[3]*vel[3])
    return speed
end

--horn functions
local function GetVhVox(vehicle)
    local vc = 0
    for i,v in pairs(GetBodyShapes(GetVehicleBody(vehicle))) do
        vc = vc + GetShapeVoxelCount(v)
    end
    return vc
end

local function GetHornType(vehicle)
    local vc = GetVhVox(vehicle)
    
    if HasTag(vehicle, "boat") then
        if vc >= 80000 then
            return "water_large"
        elseif vc >= 15000 then
            return "water_medium"
        else
            return "water_small"
        end
    else
        if vc >= 12000 then
            return "land_large"
        elseif vc >= 3100 then
            return "land_medium"
        else
            return "land_small"
        end
    end
end

--init
function init()
	DefaultData(true)
	
    --global
	global = {}
	global.debug =          false
	global.deltaFrameTime =	0
	global.lastFrameTime = 	0
    global.shapes =         FindShapes(nil,true)
    global.bodies =         FindBodies(nil,true)
    global.staticBodies =   {}
    global.fly_enabled =    GetBool("savegame.mod.global.fly.enabled")
    global.sprint_enabled = GetBool("savegame.mod.global.sprint.enabled")
    global.hands_enabled =  GetBool("savegame.mod.global.hands.enabled")
    global.phys_enabled =   GetBool("savegame.mod.global.phys.enabled")
    global.horn_enabled =   GetBool("savegame.mod.global.horn.enabled")
    global.debris_enabled = GetBool("savegame.mod.global.debris.enabled")
    global.speedmtr_enabled = GetBool("savegame.mod.global.speedmtr.enabled")
    global.sidebar =        GetBool("savegame.mod.global.sideinfo")

    for i, v in pairs(global.bodies) do
		if not IsBodyDynamic(v) then
			table.insert(global.staticBodies, v)
		end
    end

    if global.debug then
        DefaultData(false)
    end

	--fly
    --thanks to wsung and kmc for their effort on previous mods
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2416253872
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2419333267
    --thanks to Cr4xy for finding the magic number
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2580322132
    --thanks to Fall Guy 2106 for their noclip angle/rotation code
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2406955331
	fly = {}
	fly.FlyTiming = 		GetInt("savegame.mod.fly.FlyTiming")
	fly.DownKey = 			GetString("savegame.mod.fly.DownKey")
	fly.SpeedKey = 			GetString("savegame.mod.fly.SpeedKey")
	fly.NoclipKey = 		GetString("savegame.mod.fly.NoclipKey")
	fly.FreecamKey = 		GetString("savegame.mod.fly.FreecamKey")
	fly.camBaseSpeed =		GetInt("savegame.mod.fly.camBaseSpeed")
	fly.flyingMod = 		GetInt("savegame.mod.fly.flyingMod")/100
    fly.preciseMod = 		GetInt("savegame.mod.fly.preciseMod")/100
	fly.flyBaseSpeed =		GetInt("savegame.mod.fly.flyBaseSpeed")
    fly.precisionKey =      GetString("savegame.mod.fly.PreciseKey")
    fly.sandboxOnly =       GetBool("savegame.mod.fly.sandboxOnly")
    fly.healWhenDown =      GetBool("savegame.mod.fly.healWhenDown")
    fly.doDoubleTap =       GetBool("savegame.mod.fly.doDoubleTap")
    fly.flykey =            GetString("savegame.mod.fly.flykey")

	fly.freecamPos = 		Vec(0, 0, 0)
	fly.input =				nil
	fly.state = 			"idle"
	fly.spaceFrameCount	=	0
	fly.SpaceCount	=		0
    fly.WasFlying =         false
    fly.PlayerToCamDiff =   0
    fly.ncAngle =           nil

    if (not GetBool("level.sandbox")) and (fly.sandboxOnly) then
        global.fly_enabled = false
    end

    --hands
    hands = {}
    hands.toolgroup =       GetInt("savegame.mod.hands.toolgroup")

    if global.hands_enabled then
        RegisterTool("hands", "Hands", "MOD/vox/Void.vox", hands.toolgroup)
	    SetBool("game.tool.hands.enabled", true)
    end

    --sprint & walk
    --thanks to Cheejins for his move function
    --https://steamcommunity.com/sharedfiles/filedetails/?id=2415158477
    sprint = {}
    sprint.limit =          GetBool("savegame.mod.sprint.limit")
    sprint.time =           GetInt("savegame.mod.sprint.time")
    sprint.disableBT =      GetBool("savegame.mod.sprint.disableBT")
    sprint.sprintkey =      GetString("savegame.mod.sprint.sprintkey")
    sprint.speed =          GetInt("savegame.mod.sprint.speed")
    sprint.walkspeed =      GetInt("savegame.mod.sprint.walkspeed")/100
    sprint.walkKey =        GetString("savegame.mod.sprint.walkKey")
    
    sprint.Left =           sprint.time
    sprint.isBarUp =        false
    sprintfade =            0
    sprint.sprinting =      false
    sprint.regen =          1
    sprint.delay =          1.5
    sprint.lastRanOut =     0

    --pause physics
    physics = {}
    physics.toggleKB =      GetString("savegame.mod.physics.toggleKB")
    physics.easeTrans =     GetBool("savegame.mod.physics.easeTrans")
    physics.cancelMvmnt =   GetBool("savegame.mod.physics.cancelMvmnt")
    physics.enforce =       GetBool("savegame.mod.physics.enforce")
    physics.bulletdest =    GetInt("savegame.mod.physics.bulletdest")/100
    physics.smartUpdate =   GetBool("savegame.mod.physics.smartUpdate")
    physics.passiveBodies = {}

    physics.transamnt =     1
    physics.instrans =      false
    physics.state =         true

    --horn
    horn = {}
    horn.hornkey =          GetString("savegame.mod.horn.hornkey")
    horn.fadespeed =        GetInt("savegame.mod.horn.fadespeed")
    
    horn.ishorn =           false
    horn.type =             nil
    horn.fadedir =          true --false = in, true = out
    horn.fade =             0
    hornsounds = {}
    hornsounds.land_small = LoadLoop("MOD/sound/land_small.ogg", 10)
    hornsounds.land_medium = LoadLoop("MOD/sound/land_medium.ogg", 25)
    hornsounds.land_large = LoadLoop("MOD/sound/land_large.ogg", 50)
    hornsounds.water_small = LoadLoop("MOD/sound/water_small.ogg", 10)
    hornsounds.water_medium = LoadLoop("MOD/sound/water_medium.ogg", 50)
    hornsounds.water_large = LoadLoop("MOD/sound/water_large.ogg", 125)
    vhdata = {}
    vhdata.vehicles = FindVehicles(nil,true)
    vhdata.type = {}
    for i,v in pairs(vhdata.vehicles) do
        vhdata.type[v] = GetHornType(v)
    end

    --debris
    debris = {}
    debris.terminalScreen = FindScreen("terminal", true)
    debris.terminalBody = GetShapeBody(GetScreenShape(debris.terminalScreen))
    debris.count = 0

    --speedometer
    speedmtr = {}
    speedmtr.kph = GetBool("savegame.mod.speedmtr.kph")
    speedmtr.lim = GetInt("savegame.mod.speedmtr.lim")
end

--fly functions
local function holdPlayerStill()
	local curvec = GetPlayerVelocity()
	local modifiedvec = {curvec[1],curvec[2],curvec[3]}
	local factor = 200/100
	local reduxfactor = 50/100
	if curvec[2] < 0 then
		modifiedvec[2] = 0.16663932800293
	else
		modifiedvec[2] = modifiedvec[2]*reduxfactor
	end
	modifiedvec[1] = modifiedvec[1]*reduxfactor
	modifiedvec[3] = modifiedvec[3]*reduxfactor
	SetPlayerVelocity(Vec(modifiedvec[1], 0.16663932800293, modifiedvec[3]))
end

local function syncCamAndPlayer(init, tools)
    if (init) then
        fly.PlayerToCamDiff = GetPlayerCameraTransform().pos[2]-GetPlayerTransform().pos[2]
        if tools then
            fly.freecamPos = GetPlayerTransform().pos
        else
            fly.freecamPos = GetPlayerCameraTransform().pos
        end
    else
        SetPlayerTransform(Transform(Vec(fly.freecamPos[1], fly.freecamPos[2]-fly.PlayerToCamDiff, fly.freecamPos[3]), GetCameraTransform(true).rot), true)
    end
end

local function freeCam(tools)
    local cameraFront = Vec(0, 0, 0)
    local cameraUp = Vec(0, 1, 0)
	
    local camSpeed = fly.input.camSpeed * global.deltaFrameTime
    local cameraFront = getCameraFront()

    if (InputDown("up")) then
        fly.freecamPos = VecAdd(fly.freecamPos, VecScale(cameraFront, camSpeed))
    end
    if (InputDown("down")) then
        fly.freecamPos = VecSub(fly.freecamPos, VecScale(cameraFront, camSpeed))
    end
    if (InputDown("left")) then
        fly.freecamPos = VecSub(fly.freecamPos, VecScale(VecNormalize(VecCross(cameraFront, cameraUp)), camSpeed))
    end
    if (InputDown("right")) then
        fly.freecamPos = VecAdd(fly.freecamPos, VecScale(VecNormalize(VecCross(cameraFront, cameraUp)), camSpeed))
    end
    if (InputDown("space")) then
        fly.freecamPos = VecAdd(fly.freecamPos, Vec(0, camSpeed, 0))
    end
    if (InputDown(fly.DownKey)) then
        fly.freecamPos = VecSub(fly.freecamPos, Vec(0, camSpeed, 0))
    end

    local lookat = QuatLookAt(fly.freecamPos, VecAdd(fly.freecamPos, cameraFront))

    if tools then
        local dx, dy = InputValue("mousedx"), InputValue("mousedy")
        local sensitivity = GetInt("options.input.sensitivity")/100

        fly.ncAngle[1] = math.min(math.max(fly.ncAngle[1] - dy / 10 * sensitivity,-80), 80)
        fly.ncAngle[2] = fly.ncAngle[2] - dx / 10 * sensitivity
        lookat = QuatEuler(fly.ncAngle[1], fly.ncAngle[2], 0)
        SetPlayerTransform(Transform(fly.freecamPos, lookat), true)
    else
        SetCameraTransform(Transform(fly.freecamPos, lookat))
    end
end

local function flyfunc()
    if (fly.input.moving) then
        local playerTransform = GetPlayerTransform()
        local parent = TransformToParentPoint(playerTransform, fly.input.direction)
        local reverse = VecNormalize(VecSub(parent, playerTransform.pos))
        local velocity = VecScale(reverse, fly.input.flyingVel)
        local tempvely = velocity[2]
        if tempvely == 0 then
            tempvely = 0.16663932800293
        end
        SetPlayerVelocity(Vec(velocity[1], tempvely, velocity[3]))
    else
        holdPlayerStill()
    end
end

local function flyCheck()
    local input = {}

    input.direction = Vec(0, 0, 0)

    input.moving = true
	input.flyingVel = fly.flyBaseSpeed 
    input.camSpeed = fly.camBaseSpeed

    if (InputDown("up")) then
        input.direction = VecAdd(input.direction, Vec(0, 0, -1))
    end

    if (InputDown("left")) then
        input.direction = VecAdd(input.direction, Vec(-1, 0, 0))
    end

    if (InputDown("down")) then
        input.direction = VecAdd(input.direction, Vec(0, 0, 1))
    end

    if (InputDown("right")) then
        input.direction = VecAdd(input.direction, Vec(1, 0, 0))
    end

    if (InputDown("space")) then
        input.direction = VecAdd(input.direction, Vec(0, 1, 0))
    end

    if (InputDown(fly.DownKey)) then
        input.direction = VecAdd(input.direction, Vec(0, -1, 0))
    end

    if InputDown(fly.SpeedKey) then
        input.flyingVel = fly.flyBaseSpeed * fly.flyingMod
        input.camSpeed = fly.camBaseSpeed * fly.flyingMod
    end
    if InputDown(fly.precisionKey) then
        input.flyingVel = fly.flyBaseSpeed * fly.preciseMod
        input.camSpeed = fly.camBaseSpeed * fly.preciseMod
    end

    if (input.direction[1] == 0 and input.direction[2] == 0 and input.direction[3] == 0) then
        input.moving = false
    end

    return input
end

--sprint and walk functions
local function MovePlayer(speed, dir)
    local mvDir = VecScale(dir, speed)

    local camTr = GetPlayerCameraTransform()
    local camFwd = TransformToParentPoint(camTr, Vec(0,0,-1))
    camTr.pos[2] = 0
    camFwd[2] = 0

    local camSubFlatY = VecNormalize(VecSub(camTr.pos, camFwd))
    camTr.rot = QuatLookAt(GetPlayerTransform(), camSubFlatY)

    local sVel = GetPlayerVelocity()
    sVel = VecSub(camTr.pos, TransformToParentPoint(camTr, mvDir))
    sVel[2] = GetPlayerVelocity()[2]
    SetPlayerVelocity(sVel)
    -- printVec(GetPlayerVelocity())
end

--tick
function tick(dt)
    --global stuff
    local FrameTime = GetTime()
    global.deltaFrameTime = FrameTime - global.lastFrameTime
    global.lastFrameTime = FrameTime

	--fly stuff
    fly.input = flyCheck()

	if InputPressed(fly.flykey) and (global.fly_enabled) then
        if fly.doDoubleTap then
            fly.SpaceCount =  fly.SpaceCount + 1
        else
            fly.SpaceCount =  2
        end
    end
    if InputPressed(fly.NoclipKey) and (fly.state ~= "freecam") and (GetPlayerVehicle() == 0) and (global.fly_enabled) then
        if (fly.state == "idle") or (fly.state == "flying") then
            fly.WasFlying = (fly.state == "flying")
            syncCamAndPlayer(true, true)
            fly.ncAngle = Vec(toEuler(GetCameraTransform().rot))
            fly.state = "noclip"
        else
            if fly.WasFlying then
                fly.state = "flying"
            else
                fly.state = "idle"
            end
        end
    end
    if InputPressed(fly.FreecamKey) and (fly.state ~= "noclip") and (GetPlayerVehicle() == 0) and (global.fly_enabled) then
        if (fly.state == "idle") or (fly.state == "flying") then
            fly.WasFlying = (fly.state == "flying")
            syncCamAndPlayer(true, false)
            fly.state = "freecam"
        else
            if fly.WasFlying then
                fly.state = "flying"
            else
                fly.state = "idle"
            end
        end
    end

    if (fly.state == "noclip") or (fly.state == "freecam") then
	    freeCam((fly.state == "noclip"))
        holdPlayerStill()
        SetPlayerHealth(1)
	end

    if (fly.healWhenDown) and (InputDown(fly.DownKey)) and (not InputDown("space")) and (fly.state == "flying") then
        SetPlayerHealth(1)
    end

    --sprint
    if InputDown(sprint.sprintkey) and (GetPlayerVehicle() == 0) and (fly.state == "idle") and ((sprint.Left > 0) or (not sprint.limit)) and (global.sprint_enabled) then
        sprint.sprinting = true
        sprint.lastRanOut = GetTime()
        if fly.input.moving then
            MovePlayer(sprint.speed, fly.input.direction)
            sprint.Left = sprint.Left - global.deltaFrameTime
        end
    else
        sprint.sprinting = false
        if (sprint.Left < sprint.time) and (GetTime() > (sprint.lastRanOut + sprint.delay)) then
            sprint.Left = sprint.Left + sprint.regen * global.deltaFrameTime
        elseif sprint.Left > sprint.time then
            sprint.Left = sprint.time
        end
    end
    if InputDown(sprint.walkKey) and (GetPlayerVehicle() == 0) and (fly.state == "idle") and (global.sprint_enabled) then
        if fly.input.moving then
            MovePlayer(sprint.walkspeed, fly.input.direction)
        end
    end

    --pause physics
    if InputPressed(physics.toggleKB) and (not physics.instrans) and (global.phys_enabled) then
        physics.state = not physics.state
        if (physics.easeTrans) then
            physics.instrans = true
        else
            if physics.smartUpdate then
                physics.passiveBodies = {}
            end
            for i,v in pairs(global.bodies) do
                if not physics.state then
                    if physics.smartUpdate and (not IsBodyActive(v)) then
                        physics.passiveBodies[v] = true
                    end
                    SetBodyDynamic(v,false)
                else
                    if not has_value(global.staticBodies, v) then
                        SetBodyDynamic(v,true);
                        if not physics.cancelMvmnt then
                            local com = GetBodyCenterOfMass(v)
                            local worldPoint = TransformToParentPoint(GetBodyTransform(v), com)
                            if (physics.smartUpdate) and (physics.passiveBodies[v] ~= true) then
                                ApplyBodyImpulse(v, worldPoint, imp)
                            elseif (not physics.smartUpdate) then
                                ApplyBodyImpulse(v, worldPoint, imp)
                            end
                        end
                    end
                end
            end
        end
    end
    if (physics.instrans) then
        if not physics.state then
            if (physics.transamnt > physics.bulletdest) then
                SetTimeScale(physics.transamnt)
                physics.transamnt = physics.transamnt - global.deltaFrameTime/2
            else
                physics.instrans = false
                physics.transamnt = physics.bulletdest
                SetTimeScale(physics.transamnt)
            end
        else
            if (physics.transamnt < 1) then
                SetTimeScale(physics.transamnt)
                physics.transamnt = physics.transamnt + global.deltaFrameTime/2
            else
                physics.instrans = false
                physics.transamnt = 1
                SetTimeScale(physics.transamnt)
            end
        end
    end
    if (physics.transamnt < 1) and physics.easeTrans then
        SetTimeScale(physics.transamnt)
    end

    if (not physics.state) and (physics.enforce) and (not physics.easeTrans) then
        for i,v in pairs(global.bodies) do
            SetBodyDynamic(v,false)
        end
    end

    --horn
    if InputDown(horn.hornkey) and (GetPlayerVehicle() ~= 0) and (global.horn_enabled) then
        if not horn.ishorn then
            horn.type = vhdata.type[GetPlayerVehicle()]
            horn.fadedir = false --false = in, true = out
            horn.fade = 0
            horn.ishorn = true
        end
    elseif (horn.ishorn) then
        horn.fadedir = true --false = in, true = out
    end

    if (not horn.fadedir) and (horn.fade < 1) then
        horn.fade = horn.fade + global.deltaFrameTime*horn.fadespeed
    elseif (horn.fadedir) and (horn.fade > 0) then
        horn.fade = horn.fade - global.deltaFrameTime*horn.fadespeed
    end
    if horn.fade>1 then horn.fade = 1 end
    if horn.fade<0 then horn.fade = 0 end
    if horn.fade==0 and horn.ishorn then horn.ishorn = false end
    
    if horn.ishorn then
        PlayLoop(hornsounds[horn.type], GetVehicleTransform(GetPlayerVehicle()).pos, horn.fade)
    end

    --debris
    if global.debris_enabled and physics.state then
        for i,v in pairs(global.bodies) do
            local center = TransformToParentPoint(GetBodyTransform(v), GetBodyCenterOfMass(v)) 
            if (center[2] <= -19) and (HasTag(v, "tool") == false) and (HasTag(v, "valuable") == false) and (HasTag(v, "interact") == false) and (v ~= debris.terminalBody) and (v ~= GetToolBody()) then
                Delete(v)
                --DebugCross(center, 1,0,1,1)
                --DrawBodyOutline(v,1,0,0,1)
                debris.count = debris.count + 1
            end
        end
    end
end

--update
function update(dt)
	--global stuff
    if global.debug then
		DebugWatch("global.deltaFrameTime", global.deltaFrameTime)
		DebugWatch("global.lastFrameTime", global.lastFrameTime)
	end
    global.shapes = FindShapes(nil,true)
    global.bodies = FindBodies(nil,true)
    
    --fly stuff
	if (fly.state == "flying") then
        flyfunc()
    end

    if (fly.SpaceCount >= 1) then
        fly.spaceFrameCount = fly.spaceFrameCount + 1
    end

    if (fly.spaceFrameCount > fly.FlyTiming) then
        fly.spaceFrameCount = 0
        fly.SpaceCount = 0
    end

    if (fly.SpaceCount >= 2) and (GetPlayerVehicle() == 0) then
		if (fly.state == "flying") then
			fly.state = "idle"
			holdPlayerStill()
		elseif (fly.state == "noclip") or (fly.state == "freecam") then
			
		else
			fly.state = "flying"
		end
		fly.spaceFrameCount = 0
        fly.SpaceCount = 0
    end
	
	if global.debug then
		DebugWatch("fly.freecamPos", fly.freecamPos)
		DebugWatch("fly.input.moving", fly.input.moving)
		DebugWatch("fly.input", fly.input)
		DebugWatch("fly.state", fly.state)
		DebugWatch("fly.spaceFrameCount", fly.spaceFrameCount)
		DebugWatch("fly.SpaceCount", fly.SpaceCount)
        DebugWatch("fly.input.flyingVel", fly.input.flyingVel)
	end

    --sprint
    if global.debug then
		DebugWatch("sprint.isBarUp", sprint.isBarUp)
        DebugWatch("ishUp", (GetFloat("game.player.health") < 1))
        DebugWatch("sprint.sprintFade", sprintfade)
        DebugWatch("sprint.Left", sprint.Left)
        DebugWatch("target", sprint.lastRanOut + sprint.delay)
        DebugWatch("cur", GetTime())
        DebugWatch("shouldregen", (GetTime() > (sprint.lastRanOut + sprint.delay)))
	end

    --physics
    if global.debug then
        DebugWatch("physics.state", physics.state)
        DebugWatch("physics.bulletdest", physics.bulletdest)
        DebugWatch("physics.transamnt", physics.transamnt)
        DebugWatch("physics.instrans", physics.instrans)
    end

    --horn
    if global.debug then
        DebugWatch("horn.ishorn",horn.ishorn)
        DebugWatch("horn.type",horn.type)
        DebugWatch("horn.fadedir",horn.fadedir)
        DebugWatch("horn.fade",horn.fade)
        local vc = 0
        for i,v in pairs(GetBodyShapes(GetVehicleBody(GetPlayerVehicle()))) do
            vc = vc + GetShapeVoxelCount(v)
        end
        DebugWatch("vehicle voxcount", vc)
    end
end

--draw
function draw()
	--Sprint & walk stuff
    if sprint.disableBT and (global.sprint_enabled) then
        SetBool("game.tool.steroid.enabled", false)
    end

    if (sprint.sprinting or (sprint.Left ~= sprint.time)) and (sprint.limit) and (global.sprint_enabled) and (GetPlayerVehicle() == 0) then
        local offset = 0
        local cur = sprint.Left/sprint.time
        local ishUp = (GetFloat("game.player.health") < 1)
        if sprintfade ~= 0 then
            offset = 35
        end
        if (ishUp) and (not sprint.isBarUp) then
            SetValue("sprintfade", 1, "easeout", 0.5)
        elseif (not ishUp) and (sprint.isBarUp) then
            SetValue("sprintfade", 0, "easein", 0.5)
        end
        sprint.isBarUp = ishUp

        UiPush() --Taken from hud.lua at line 1354
            UiTranslate(UiWidth() - 144, UiHeight() - (44 + offset*sprintfade))

            UiColor(0,0,0, 0.5)
            UiPush()
                UiColor(1,1,1)
                UiFont("bold.ttf", 24)
                UiTranslate(0, 22)
                if cur < 0.1 then
                    if math.mod(GetTime(), 1.0) < 0.5 then
                        UiColor(1, 0, 0,  1.0)
                    else
                        UiColor(1, 0, 0,  0.1)
                    end
                elseif cur < 0.5 then
                    UiColor(1, 0, 0)
                end
                UiAlign("right")
                UiText("SPRINT")
            UiPop()

            UiTranslate(10, 4)
            local w = 110
            local h = 20
            UiPush()
                UiAlign("left top")
                UiColor(0, 0, 0, 0.5)
                UiImageBox("ui/common/box-solid-10.png", w, h, 6, 6)
                if cur > 0 then
                    UiTranslate(2, 2)
                    w = (w-4)*cur
                    if w < 12 then w = 12 end
                    h = h-4
                    UiColor(1,cur*2,cur,1)
                    UiImageBox("ui/common/box-solid-6.png", w, h, 6, 6)
                end
            UiPop()

        UiPop()
    end

    --horn in vehicle controls
    if (GetPlayerVehicle() ~= 0) and (global.horn_enabled) then
        UiPush()
            local offset = 0
            if (GetFloat("game.player.health") < 1) then
                offset = 50
            end
            --hud.lua line 714
            local vehicle = GetPlayerVehicle()
            local movement = string.upper(GetString("game.input.up")).." "..string.upper(GetString("game.input.left")).." "..string.upper(GetString("game.input.down")).." "..string.upper(GetString("game.input.right"))
            local info = {}
            if HasTag(vehicle, "cranetop") then
                info[#info+1] = {movement, "Turn/extend"}
                info[#info+1] = {"LMB RMB", "Raise/lower"}
                info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Hook"}
            else
                info[#info+1] = {movement, "Drive"}
                if HasTag(vehicle, "cranebottom") then
                    info[#info+1] = {"LMB RMB", "Support legs"}
                elseif HasTag(vehicle, "crane") then
                    info[#info+1] = {"LMB RMB", "Arm"}
                    info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Hook"}
                elseif HasTag(vehicle, "dumptruck") then
                    info[#info+1] = {"LMB RMB", "Bed"}
                    info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Brake"}
                elseif HasTag(vehicle, "frontloader") then
                    info[#info+1] = {"LMB RMB", "Shovel"}
                    info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Brake"}
                elseif HasTag(vehicle, "skylift") then
                    info[#info+1] = {"LMB RMB", "Lift"}
                    info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Brake"}
                elseif HasTag(vehicle, "forklift") then
                    info[#info+1] = {"LMB RMB", "Fork"}
                    info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Brake"}
                elseif HasTag(vehicle, "tank") then
                    info[#info+1] = {"LMB", "Fire"}
                elseif HasTag(vehicle, "boat") then
                else
                    info[#info+1] = {string.upper(GetString("game.input.handbrake")), "Handbrake"}
                end
            end
            info[#info+1] = {string.upper(GetString("game.input.interact")), "Exit vehicle"}

            local kw = 0
            local vw = 0
            for i=1, #info do
                UiFont("bold.ttf", 22)
                local w0 = UiGetTextSize(info[i][1])
                UiFont("regular.ttf", 22)
                local w1 = UiGetTextSize(info[i][2])
                kw = math.max(kw,w0)
                vw = math.max(vw,w1)
            end

            local w = kw + vw + 50
            local h = #info*22 + 30
            local ah = 42
            UiAlign("top left")
            UiTranslate(UiWidth()-w-20, UiHeight()-ah-20-h-10-offset)
            UiColor(0+(horn.fade),0+(horn.fade),0+(horn.fade),0.5)
            UiImageBox("ui/common/box-solid-6.png", w, ah, 6, 6)
            UiTranslate(kw + 20, 28)
            UiColor(1-(horn.fade),1-(horn.fade),1-(horn.fade))
            UiFont("bold.ttf", 22)
            UiAlign("right")
            UiText(horn.hornkey)
            UiTranslate(10, 0)
            UiFont("regular.ttf", 22)
            UiAlign("left")
            UiText("Horn")
        UiPop()
    end
    
    --speedometer above vh hp
    if (global.speedmtr_enabled) and (GetPlayerVehicle() ~= 0) then
        UiPush()
		    UiFont("bold.ttf", 20)
            UiTranslate(UiCenter(), UiHeight()-82)
            local maxspd = speedmtr.lim
            local mult = 3.6
            local unit = " KM/H"
            if not speedmtr.kph then
                maxspd = maxspd * 0.6214
                mult = 2.23694
                unit = " MPH"
            end
            local airspeed = getBodySpeed(GetVehicleBody(GetPlayerVehicle()))*mult
            local speed = airspeed/maxspd
            if speed > 1 then speed = 1 end
            UiTranslate(-100, 0)
            progressBar(200, 20, speed)
            UiTranslate(100, -12)
            if speed > 0.9 then
                if math.mod(GetTime(), 1.0) < 0.5 then
                    UiColor(1, 0, 0,  1.0)
                else
                    UiColor(1, 0, 0,  0.1)
                end
            elseif speed > 0.5 then
                UiColor(1, 0, 0)
            end
            UiAlign("center middle")
		    UiText(round(airspeed, 0)..unit)
	    UiPop()
    end

    --Sidebar for pause physics and flight
    if global.sidebar then
        local DispList = {
            {title="Artzert's Utilities", icon="MOD/image/a-utils_smallest.png"}
        }
        local ftb
    
        if global.fly_enabled then
            --flight
            ftb = {}
            ftb.title = fly.state:gsub("^%l", string.upper)
            ftb.icon = "MOD/image/flight_smallest.png"
            table.insert(DispList, ftb)
        end
        
        if global.phys_enabled then
            --pause physics
            ftb = {}
            ftb.title = "Realtime"
            ftb.icon = "MOD/image/play_smallest.png"
            if not physics.state then
                if physics.easeTrans then
                    ftb.icon = "MOD/image/bullet_smallest.png"
                    ftb.title = "Bullet Time!"
                else
                    ftb.icon = "MOD/image/pausephysics_smallest.png"
                    ftb.title = "Paused"
                end
            end
            table.insert(DispList, ftb)
        end

        if global.debris_enabled then
            --debris
            ftb = {}
            ftb.title = debris.count .." Removed"
            ftb.icon = "MOD/image/debris_smallest.png"
            table.insert(DispList, ftb)
        end
    
        UiPush()
            UiTranslate(15,15)
            for i,v in pairs(DispList) do
                UiAlign("left top")
                if i~=1 then
                    UiColor(0,0,0,0.6)
                else
                    UiColor(1,1,1,0.6)
                end
                UiImageBox("ui/common/box-solid-6.png", 160, 26, 10, 10)
                UiPush()
                    if i~=1 then
                        UiColor(1,1,1,0.6)
                    else
                        UiColor(0,0,0,0.6)
                    end
                    UiAlign("left middle")
                    UiTranslate(8,13)
                    UiImage(v.icon)
                    UiFont("regular.ttf", 15)
                    UiTranslate(18,0)
                    UiText(v.title)
                UiPop()
                UiTranslate(0,31)
            end
        UiPop()
    end
end